"""Query the `GenomicArrayDataset`.

This class provides methods to access the directory containing the
generated TileDB files usually using the
:py:func:`~genomicarray.build_genomicarray.build_genomicarray`.

Example:

    .. code-block:: python

        from genomicarray import GenomicArrayDataset

        garr = GenomicArrayDataset(dataset_path="/path/to/genomicarray/dir")
        result1 = garr[0:10, 0]

        print(result1)
"""

import os
from typing import List, Sequence, Union

import pandas as pd
import tiledb

from . import queryutils_tiledb_frame as qtd
from .GenomicArrayDatasetSlice import GenomicArrayDatasetSlice

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


class GenomicArrayDataset:
    """A class that represent a collection of features and their associated coverage in a TileDB backed store."""

    def __init__(
        self,
        dataset_path: str,
        matrix_tdb_uri: str = "coverage",
        feature_annotation_uri: str = "feature_annotation",
        sample_metadata_uri: str = "sample_metadata",
    ):
        """Initialize a ``GenomicArrayDataset``.

        Args:
            dataset_path:
                Path to the directory containing the TileDB stores.
                Usually the ``output_path`` from the
                :py:func:`~genomicarray.build_genomicarray.build_genomicarray`.

            matrix_tdb_uri:
                Relative path to matrix store.

            feature_annotation_uri:
                Relative path to feature annotation store.

            sample_metadata_uri:
                Relative path to sample metadata store.
        """

        if not os.path.isdir(dataset_path):
            raise ValueError("'dataset_path' is not a directory.")

        self._dataset_path = dataset_path
        # TODO: Maybe switch to on-demand loading of these objects
        self._matrix_tdb_tdb = tiledb.open(f"{dataset_path}/{matrix_tdb_uri}", "r")
        self._feature_annotation_tdb = tiledb.open(f"{dataset_path}/{feature_annotation_uri}", "r")
        self._sample_metadata_tdb = tiledb.open(f"{dataset_path}/{sample_metadata_uri}", "r")

    def __del__(self):
        self._matrix_tdb_tdb.close()
        self._feature_annotation_tdb.close()
        self._sample_metadata_tdb.close()

    ####
    ## Subset methods for the `feature_annotation` TileDB file.
    ####
    def get_feature_annotation_columns(self) -> List[str]:
        """Get annotation column names from ``feature_annotation`` store.

        Returns:
            List of available annotations.
        """
        return qtd.get_schema_names_frame(self._feature_annotation_tdb)

    def get_feature_annotation_column(self, column_name: str) -> pd.DataFrame:
        """Access a column from the ``feature_annotation`` store.

        Args:
            column_name:
                Name of the column or attribute. Usually one of the column names
                from of :py:meth:`~get_feature_annotation_columns`.

        Returns:
            A list of values for this column.
        """
        res = qtd.get_a_column(self._feature_annotation_tdb, column_name=column_name)
        return res[column_name]

    def get_feature_annotation_index(self) -> List[str]:
        """Get index of the ``feature_annotation`` store.

        Returns:
            List of feature ids.
        """
        res = qtd.get_a_column(self._feature_annotation_tdb, "genarr_feature_index")
        return res["genarr_feature_index"].tolist()

    def get_feature_subset(self, subset: Union[slice, List[str], tiledb.QueryCondition], columns=None) -> pd.DataFrame:
        """Slice the ``feature_annotation`` store.

        Args:
            subset:
                A list of integer indices to subset the ``feature_annotation``
                store.

                Alternatively, may provide a
                :py:class:`tiledb.QueryCondition` to query the store.

                Alternatively, may provide a list of strings to match with
                the index of ``feature_annotation`` store.

            columns:
                List of specific column names to access.

                Defaults to None, in which case all columns are extracted.

        Returns:
            A pandas Dataframe of the subset.
        """

        if isinstance(columns, str):
            columns = [columns]

        if columns is None:
            columns = self.get_feature_annotation_columns()
        else:
            _not_avail = []
            for col in columns:
                if col not in self.get_feature_annotation_columns():
                    _not_avail.append(col)

            if len(_not_avail) > 0:
                raise ValueError(f"Columns '{', '.join(_not_avail)}' are not available.")

        if qtd._is_list_strings(subset):
            subset = self._get_indices_for_gene_list(subset)

        return qtd.subset_frame(self._feature_annotation_tdb, subset=subset, columns=columns)

    ####
    ## Subset methods for the `sample_metadata` TileDB file.
    ####
    def get_sample_metadata_columns(self) -> List[str]:
        """Get column names from ``sample_metadata`` store.

        Returns:
            List of available metadata columns.
        """
        return qtd.get_schema_names_frame(self._sample_metadata_tdb)

    def get_sample_metadata_column(self, column_name: str) -> pd.DataFrame:
        """Access a column from the ``sample_metadata`` store.

        Args:
            column_name:
                Name of the column or attribute. Usually one of the column names
                from of :py:meth:`~get_sample_metadata_columns`.

        Returns:
            A list of values for this column.
        """
        res = qtd.get_a_column(self._sample_metadata_tdb, column_name=column_name)
        return res[column_name]

    def get_sample_subset(self, subset: Union[slice, tiledb.QueryCondition], columns=None) -> pd.DataFrame:
        """Slice the ``sample_metadata`` store.

        Args:
            subset:
                A list of integer indices to subset the ``sample_metadata``
                store.

                Alternatively, may also provide a
                :py:class:`tiledb.QueryCondition` to query the store.

            columns:
                List of specific column names to access.

                Defaults to None, in which case all columns are extracted.

        Returns:
            A pandas Dataframe of the subset.
        """
        if isinstance(columns, str):
            columns = [columns]

        if columns is None:
            columns = self.get_sample_metadata_columns()
        else:
            _not_avail = []
            for col in columns:
                if col not in self.get_sample_metadata_columns():
                    _not_avail.append(col)

            if len(_not_avail) > 0:
                raise ValueError(f"Columns '{', '.join(_not_avail)}' are not available.")

        return qtd.subset_frame(self._sample_metadata_tdb, subset=subset, columns=columns)

    ####
    ## Subset methods for the `matrix` TileDB file.
    ####
    def get_matrix_subset(self, subset: Union[int, Sequence, tuple]) -> pd.DataFrame:
        """Slice the ``matrix`` store.

        Args:
            subset:
                Any `slice`supported by TileDB's array slicing.
                For more info refer to
                <TileDB docs https://docs.tiledb.com/main/how-to/arrays/reading-arrays/basic-reading>_.

        Returns:
            A pandas Dataframe of the subset.
        """
        if isinstance(subset, (str, int)):
            return qtd.subset_array(
                self._matrix_tdb_tdb,
                subset,
                slice(None),
                shape=(len(subset), self.shape[1]),
            )

        if isinstance(subset, tuple):
            if len(subset) == 0:
                raise ValueError("At least one slicing argument must be provided.")

            if len(subset) == 1:
                return qtd.subset_array(
                    self._matrix_tdb_tdb,
                    subset[0],
                    slice(None),
                    shape=(len(subset[0]), self.shape[1]),
                )
            elif len(subset) == 2:
                return qtd.subset_array(
                    self._matrix_tdb_tdb,
                    subset[0],
                    subset[1],
                    shape=(len(subset[0]), len(subset[1])),
                )
            else:
                raise ValueError(f"`{type(self).__name__}` only supports 2-dimensional slicing.")

    ####
    ## Subset methods by cell and gene dimensions.
    ####
    def get_slice(
        self,
        feature_subset: Union[slice, int],
        sample_subset: Union[slice, List[str], tiledb.QueryCondition],
    ) -> GenomicArrayDatasetSlice:
        """Subset a ``GenomicArrayDataset``.

        Args:
            sample_subset:
                Integer indices, a boolean filter, or (if the current object is
                named) names specifying the columns (or samples) to retain.

            feature_subset:
                Integer indices, a boolean filter, or (if the current object is
                named) names specifying the rows (or features/genes) to retain.

        Returns:
            A :py:class:`~genomicarray.GenomicArrayDatasetSlice.GenomicArrayDatasetSlice` object
            containing the `sample_metadata`, `feature_annotation` and the matrix for
            the given slice ranges.
        """
        _ssubset = self.get_sample_subset(sample_subset)
        _sample_indices = _ssubset.index.tolist()

        if not isinstance(feature_subset, (int, slice)):
            raise TypeError("feature indices must be continous; either a 'slice' or 'int' index.")
        _fsubset = self.get_feature_subset(feature_subset)
        start_findex = _fsubset["genarr_feature_start_index"].astype(int).min()
        end_findex = _fsubset["genarr_feature_end_index"].astype(int).max()

        # expand intervals
        final_rows = []
        for row in _fsubset.itertuples():
            for i, _ in enumerate(range(int(row.genarr_feature_start_index), int(row.genarr_feature_end_index))):
                final_rows.append(row._replace(starts=i + row.starts, ends=i + row.starts + 1))
        _feature_df = pd.DataFrame(final_rows)

        _msubset = self.get_matrix_subset((list(range(start_findex, end_findex)), _sample_indices))

        return GenomicArrayDatasetSlice(
            _ssubset,
            _feature_df,
            _msubset,
        )

    ####
    ## Dunder method to use `[]` operator.
    ####
    def __getitem__(
        self,
        args: Union[int, Sequence, tuple],
    ) -> GenomicArrayDatasetSlice:
        """Subset a ``GenomicArrayDataset``.

        Mostly an alias to :py:meth:`~.get_slice`.

        Args:
            args:
                Integer indices, a boolean filter, or (if the current object is
                named) names specifying the ranges to be extracted.

                Alternatively a tuple of length 1. The first entry specifies
                the rows (or cells) to retain based on their names or indices.

                Alternatively a tuple of length 2. The first entry specifies
                the rows (or cells) to retain, while the second entry specifies the
                columns (or features/genes) to retain, based on their names or indices.

        Raises:
            ValueError:
                If too many or too few slices provided.

        Returns:
            A :py:class:`~genomicarray.GenomicArrayDatasetSlice.GenomicArrayDatasetSlice` object
            containing the `sample_metadata`, `feature_annotation` and the matrix.
        """
        if isinstance(args, (str, int)):
            return self.get_slice(args, slice(None))

        if isinstance(args, tuple):
            if len(args) == 0:
                raise ValueError("At least one slicing argument must be provided.")

            if len(args) == 1:
                return self.get_slice(args[0], slice(None))
            elif len(args) == 2:
                return self.get_slice(args[0], args[1])
            else:
                raise ValueError(f"`{type(self).__name__}` only supports 2-dimensional slicing.")

        raise TypeError("args must be a sequence or a scalar integer or string or a tuple of atmost 2 values.")

    ####
    ## Misc methods.
    ####
    @property
    def shape(self):
        return (
            self._feature_annotation_tdb.nonempty_domain()[0][1] + 1,
            self._sample_metadata_tdb.nonempty_domain()[0][1] + 1,
        )

    def __len__(self):
        return self.shape[0]

    ####
    ## Printing.
    ####

    def __repr__(self) -> str:
        """
        Returns:
            A string representation.
        """
        output = f"{type(self).__name__}(number_of_rows={self.shape[0]}"
        output += f", number_of_columns={self.shape[1]}"
        output += ", at path=" + self._dataset_path

        output += ")"
        return output

    def __str__(self) -> str:
        """
        Returns:
            A pretty-printed string containing the contents of this object.
        """
        output = f"class: {type(self).__name__}\n"

        output += f"number_of_rows: {self.shape[0]}\n"
        output += f"number_of_columns: {self.shape[1]}\n"
        output += f"path: '{self._dataset_path}'\n"

        return output



import os
import shutil
from typing import Union

import numpy as np
import tiledb

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def create_tiledb_array(
    tiledb_uri_path: str,
    x_dim_length: int = None,
    y_dim_length: int = None,
    x_dim_name: str = "feature_index",
    y_dim_name: str = "sample_index",
    matrix_attr_name: str = "data",
    x_dim_dtype: np.dtype = np.uint32,
    y_dim_dtype: np.dtype = np.uint32,
    matrix_dim_dtype: np.dtype = np.uint32,
    is_sparse: bool = True,
):
    """Create a TileDB file with the provided attributes to persistent storage.

    This will materialize the array directory and all
    related schema files.

    Args:
        tiledb_uri_path:
            Path to create the array TileDB file.

        x_dim_length:
            Number of entries along the x/fastest-changing dimension.
            e.g. Number of cells.
            Defaults to None, in which case, the max integer value of
            ``x_dim_dtype`` is used.

        y_dim_length:
            Number of entries along the y dimension.
            e.g. Number of genes.
            Defaults to None, in which case, the max integer value of
            ``y_dim_dtype`` is used.

        x_dim_name:
            Name for the x-dimension.
            Defaults to "feature_index".

        y_dim_name:
            Name for the y-dimension.
            Defaults to "sample_index".

        matrix_attr_name:
            Name for the attribute in the array.
            Defaults to "data".

        x_dim_dtype:
            NumPy dtype for the x-dimension.
            Defaults to np.uint32.

        y_dim_dtype:
            NumPy dtype for the y-dimension.
            Defaults to np.uint32.

        matrix_dim_dtype:
            NumPy dtype for the values in the matrix.
            Defaults to np.uint32.

        is_sparse:
            Whether the matrix is sparse.
            Defaults to True.
    """

    if x_dim_length is None:
        x_dim_length = np.iinfo(x_dim_dtype).max

    if y_dim_length is None:
        y_dim_length = np.iinfo(y_dim_dtype).max

    xdim = tiledb.Dim(name=x_dim_name, domain=(0, x_dim_length - 1), dtype=x_dim_dtype)
    ydim = tiledb.Dim(name=y_dim_name, domain=(0, y_dim_length - 1), dtype=y_dim_dtype)

    dom = tiledb.Domain(xdim, ydim)

    # expecting counts
    tdb_attr = tiledb.Attr(
        name=matrix_attr_name,
        dtype=matrix_dim_dtype,
        filters=tiledb.FilterList([tiledb.GzipFilter()]),
    )

    schema = tiledb.ArraySchema(
        domain=dom,
        sparse=is_sparse,
        attrs=[tdb_attr],
        cell_order="row-major",
        tile_order="row-major",
    )

    if os.path.exists(tiledb_uri_path):
        shutil.rmtree(tiledb_uri_path)

    tiledb.Array.create(tiledb_uri_path, schema)

    tdbfile = tiledb.open(tiledb_uri_path, "w")
    tdbfile.close()


def write_frame_intervals_to_tiledb(
    tiledb_array_uri: Union[str, tiledb.SparseArray],
    data: np.ndarray,
    y_idx: int,
    value_dtype: np.dtype = np.float32,
):
    """Append and save array data to TileDB. Expect data for one full sample (column).

    Args:
        tiledb_array_uri:
            TileDB array object or path to a TileDB object.

        data:
            numpy array to write to TileDB, must contain
            columns, "start", "end" and "value".

        value_dtype:
            NumPy dtype to reformat the matrix values.
            Defaults to ``float32``.
    """
    if data is None or len(data) == 0:
        return

    if not isinstance(data, (np.ndarray)):
        raise TypeError("Intervals not provided as numpy ndarray.")

    tiledb_fp = tiledb_array_uri
    if isinstance(tiledb_array_uri, str):
        tiledb_fp = tiledb.open(tiledb_array_uri, "w")

    tiledb_fp[0 : len(data), y_idx] = data.astype(value_dtype)
    tiledb_fp.close()


def write_array_chunks_to_tiledb(
    tiledb_array_uri: Union[str, tiledb.SparseArray],
    data: np.ndarray,
    x_idx: np.ndarray,
    y_idx: int,
    value_dtype: np.dtype = np.uint32,
):
    """Write chunks of array to the tiledb."""
    tiledb_fp = tiledb_array_uri
    if isinstance(tiledb_array_uri, str):
        tiledb_fp = tiledb.open(tiledb_array_uri, "w")

    if not isinstance(data, (np.ndarray)):
        raise TypeError("'data' is not an `ndarray`.")

    tiledb_fp[0:x_idx, y_idx] = data.astype(value_dtype)
    tiledb_fp.close()


def optimize_tiledb_array(tiledb_array_uri: str, verbose: bool = True):
    """Consolidate TileDB fragments."""
    if verbose:
        print(f"Optimizing {tiledb_array_uri}")

    frags = tiledb.array_fragments(tiledb_array_uri)
    if verbose:
        print("Fragments before consolidation: {}".format(len(frags)))

    cfg = tiledb.Config()
    cfg["sm.consolidation.step_min_frags"] = 1
    cfg["sm.consolidation.step_max_frags"] = 200
    tiledb.consolidate(tiledb_array_uri, config=cfg)
    tiledb.vacuum(tiledb_array_uri)

    frags = tiledb.array_fragments(tiledb_array_uri)
    if verbose:
        print("Fragments after consolidation: {}".format(len(frags)))



"""A dataloader using TileDB files in the pytorch-lightning framework.

This class provides a dataloader using the generated TileDB files built using the
:py:func:`~genomicarrays.build_genomicarray.build_genomicarray`.

Example:

    .. code-block:: python

        from genomicarrays.dataloader import TorchDataset

        ds = TorchDataset(
            dataset_path="/path/to/genarr/dir"
        )

        print(ds[0])
"""

import logging
from typing import Union

from torch.utils.data import Dataset

from .GenomicArrayDataset import GenomicArrayDataset

log = logging.getLogger(__name__)


__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


class TorchDataset(Dataset):
    """A class that extends pytorch :py:class:`~torch.utils.data.Dataset` to enumerate features and samples using
    TileDB."""

    def __init__(self, dataset_path: Union[str, GenomicArrayDataset]):
        """Initialize a ``gaTorchDataset``.

        Args:
            dataset_path:
                Path to the directory containing the TileDB files.

                Alternatively, may also provide a ``GenomicArrayDataset``
                object.
        """
        self._dataset = dataset_path
        if isinstance(dataset_path, str):
            self._dataset = GenomicArrayDataset(dataset_path=dataset_path)

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx):
        return self._dataset[idx, :]

    def __repr__(self) -> str:
        """
        Returns:
            A string representation.
        """
        output = f"{type(self).__name__}("
        output += f"number_of_features={self._dataset.shape[0]}"
        output += f"number_of_samples={self._dataset.shape[1]}"
        output += ")"

        return output

    def __str__(self) -> str:
        """
        Returns:
            A pretty-printed string containing the contents of this object.
        """
        output = f"class: {type(self).__name__}\n"
        output += f"number_of_features: {self._dataset.shape[0]}\n"
        output += f"number_of_samples: {self._dataset.shape[1]}\n"

        return output



from typing import Optional, Tuple

import numpy as np
import pandas as pd
import pyBigWig as bw

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def extract_bw_values(
    bw_path: str,
    chrom: str,
) -> Tuple[np.ndarray, int]:
    bwfile = bw.open(bw_path)
    if chrom not in bwfile.chroms():
        return None, None

    chrom_length = bwfile.chroms(chrom)
    data = bwfile.values(chrom, 0, chrom_length)
    return np.array(data), chrom_length


# def extract_aspd():
#     df_data = pd.DataFrame(data, columns=["start", "end", "value"])
#     df_data["chrom"] = row.chrom
#     data_nnz = df_data[df_data["value"] > 0]

#     data_nnz["tmp_group"] = (
#         data_nnz["value"] != data_nnz["value"].shift()
#     ).cumsum()
#     results.append(
#         data_nnz.groupby("tmp_group").agg(
#             {"start": "first", "end": "last", "value": "first"}
#         )
#     )


def wrapper_extract_bw_values(
    bw_path: str,
    intervals: pd.DataFrame,
    agg_func: Optional[callable],
    val_dtype: np.dtype = np.float32,
    total_length: int = None,
) -> np.ndarray:
    if total_length is None:
        total_length = len(intervals)

    if agg_func is not None:
        return extract_bw_values_as_vec(bw_path=bw_path, intervals=intervals, agg_func=agg_func, val_dtype=val_dtype)
    else:
        return extract_bw_intervals_as_vec(
            bw_path=bw_path, intervals=intervals, val_dtype=val_dtype, total_length=total_length
        )


def extract_bw_values_as_vec(
    bw_path: str,
    intervals: pd.DataFrame,
    agg_func: Optional[callable] = None,
    val_dtype: np.dtype = np.float32,
) -> np.ndarray:
    """Extract data from BigWig for a given region and apply the aggregate function.

    Args:
        bw_path:
            Path to the BigWig file.

        intervals:
            List of intervals to extract.

        agg_func:
            Aggregate function to apply.
            Defaults to None.

        val_dtype:
            Dtype of the resulting array.

    Returns:
        A vector with length as the number of intervals,
        a value if the file contains the data for the corresponding
        region or ``np.nan`` if the region is not measured.
    """
    bwfile = bw.open(bw_path)

    results = []
    for row in intervals.itertuples():
        if row.seqnames in bwfile.chroms():
            try:
                data = bwfile.values(row.seqnames, row.starts, row.ends, numpy=True)
                if data is not None and len(data) != 0:
                    results.append(agg_func(data))
                else:
                    results.append(np.nan)
            except Exception as _:
                results.append(np.nan)
        else:
            results.append(np.nan)

    return np.array(results, dtype=val_dtype)


def _get_empty_array(size, val_dtype):
    out_array = np.empty(size, dtype=val_dtype)
    out_array.fill(np.nan)

    return out_array


def extract_bw_intervals_as_vec(
    bw_path: str,
    intervals: pd.DataFrame,
    total_length: int,
    val_dtype: np.dtype = np.float32,
) -> np.ndarray:
    """Extract data from BigWig for a given region.

    Args:
        bw_path:
            Path to the BigWig file.

        intervals:
            List of intervals to extract.

        total_length:
            Size of all the regions.

        val_dtype:
            Dtype of the resulting array.

    Returns:
        A vector with length as the number of intervals,
        a value if the file contains the data for the corresponding
        region or ``np.nan`` if the region is not measured.
    """
    bwfile = bw.open(bw_path)

    out_array = _get_empty_array(total_length, val_dtype=val_dtype)

    for row in intervals.itertuples():
        if row.seqnames in bwfile.chroms():
            try:
                data = bwfile.intervals(row.seqnames, row.starts, row.ends)
                tmp_out = _get_empty_array(row.ends - row.starts, val_dtype=val_dtype)
                if data is not None and len(data) != 0:
                    for d in data:
                        _strt = max(0, d[0] - row.starts)
                        _end = min(d[1] - row.starts, row.ends)
                        tmp_out[_strt:_end] = d[2]

                out_array[row.genarr_feature_index_start : row.genarr_feature_index_start + row.ends - row.starts] = (
                    tmp_out
                )
            except Exception as _:
                pass

    return out_array



import tempfile

import anndata
import numpy as np
import pandas as pd
import pytest
import tiledb
from genomicarrays import (
    GenomicArrayDataset,
    build_genomicarray,
    FeatureAnnotationOptions,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_query_agg():
    tempdir = tempfile.mkdtemp()

    strts = np.arange(300, 600, 20)
    features = pd.DataFrame(
        {"seqnames": ["chr1"] * 15, "starts": strts, "ends": strts + 15}
    )

    dataset = build_genomicarray(
        output_path=tempdir,
        files=["tests/data/test1.bw", "tests/data/test2.bw"],
        features=features,
        genome_fasta="tests/data/test.fa",
        feature_annotation_options=FeatureAnnotationOptions(
            aggregate_function=np.nanmean
        ),
    )

    assert dataset is not None
    assert isinstance(dataset, GenomicArrayDataset)

    cd = GenomicArrayDataset(dataset_path=tempdir)

    ffp = tiledb.open(f"{tempdir}/feature_annotation", "r")

    features_rt = ffp.df[:]
    assert np.allclose(np.array(features["starts"]), np.array(features_rt["starts"]))
    assert np.allclose(np.array(features["ends"]), np.array(features_rt["ends"]))
    assert all([a == b for a, b in zip(features["seqnames"], features_rt["seqnames"])])
    assert "genarr_feature_index" in features_rt.columns

    assert cd.shape == (15, 2)
    assert len(cd) == 15

    result1 = cd[:, 0]

    assert result1 is not None
    assert result1.matrix.shape == (len(features), 1)

    assert np.allclose(
        result1.matrix.flatten(),
        np.repeat(1, 15),
    )

    result2 = cd[:, 1]
    assert result2.matrix.shape == (len(features), 1)

    expected = np.concatenate((np.repeat(0.5, 10), np.repeat(np.nan, 5)))

    assert np.allclose(
        result2.matrix.flatten()[:10],
        expected[:10],
    )

    assert cd.get_sample_metadata_columns() == [
        "genarr_sample",
    ]
    assert len(cd.get_sample_metadata_column("genarr_sample")) == 2
    assert len(cd.get_sample_subset("genarr_sample == 'sample_1'")) == 1

    assert sorted(cd.get_feature_annotation_columns()) == sorted(
        [
            "ends",
            "genarr_feature_index",
            "genarr_feature_start_index",
            "genarr_feature_end_index",
            "seqnames",
            "sequences",
            "starts",
            "widths",
        ]
    )
    assert len(cd.get_feature_annotation_column("genarr_feature_index")) == 15
    assert len(cd.get_feature_subset("genarr_feature_index == 1")) == 1

    result1 = cd.get_slice(slice(0, 5), slice(None))
    assert result1 is not None
    assert result1.matrix.shape == (6, 2)

    assert result1.to_anndata() is not None
    assert result1.to_rangedsummarizedexperiment() is not None


def test_query_noagg():
    tempdir = tempfile.mkdtemp()

    strts = np.arange(300, 600, 20)
    features = pd.DataFrame(
        {"seqnames": ["chr1"] * 15, "starts": strts, "ends": strts + 15}
    )

    dataset = build_genomicarray(
        output_path=tempdir,
        files=["tests/data/test1.bw", "tests/data/test2.bw"],
        features=features,
        genome_fasta="tests/data/test.fa",
    )

    assert dataset is not None
    assert isinstance(dataset, GenomicArrayDataset)

    cd = GenomicArrayDataset(dataset_path=tempdir)

    ffp = tiledb.open(f"{tempdir}/feature_annotation", "r")

    features_rt = ffp.df[:]
    assert np.allclose(np.array(features["starts"]), np.array(features_rt["starts"]))
    assert np.allclose(np.array(features["ends"]), np.array(features_rt["ends"]))
    assert all([a == b for a, b in zip(features["seqnames"], features_rt["seqnames"])])
    assert "genarr_feature_index" in features_rt.columns

    assert cd.shape == (15, 2)
    assert len(cd) == 15

    result1 = cd[:, 0]

    assert result1 is not None
    assert result1.matrix.shape == (225, 1)

    assert np.all(np.isnan(result1.matrix.flatten()))

    result2 = cd[:, 1]
    assert result2.matrix.shape == (225, 1)

    assert np.all(np.isnan(result2.matrix.flatten()))

    assert cd.get_sample_metadata_columns() == [
        "genarr_sample",
    ]
    assert len(cd.get_sample_metadata_column("genarr_sample")) == 2
    assert len(cd.get_sample_subset("genarr_sample == 'sample_1'")) == 1

    assert sorted(cd.get_feature_annotation_columns()) == sorted(
        [
            "ends",
            "genarr_feature_index",
            "genarr_feature_start_index",
            "genarr_feature_end_index",
            "seqnames",
            "sequences",
            "starts",
            "widths",
        ]
    )
    assert len(cd.get_feature_annotation_column("genarr_feature_index")) == 15
    assert len(cd.get_feature_subset("genarr_feature_index == 1")) == 1

    result1 = cd.get_slice(slice(0, 5), slice(None))
    assert result1 is not None
    assert result1.matrix.shape == (90, 2)

    assert result1.to_anndata() is not None
    assert result1.to_rangedsummarizedexperiment() is not None



"""Build the `GenomicArrayDatset`.

This modules provides tools for converting genomic data from BigWig format to TileDB.
It supports parallel processing for handling large collection of genomic datasets.

Example:

    .. code-block:: python

        import pyBigWig as bw
        import numpy as np
        import tempfile
        from genomicarrays import build_genomicarray, MatrixOptions

        # Create a temporary directory
        tempdir = tempfile.mkdtemp()

        # Read BigWig objects
        bw1 = bw.open("path/to/object1.bw", "r")
        # or just provide the path
        bw2 = "path/to/object2.bw"

        features = pd.DataFrame({
            "seqnames": ["chr1", "chr1"],
            "starts": [1000, 2000],
            "ends": [1500, 2500]
        })

        # Build GenomicArray
        dataset = build_genomicarray(
            features=features
            output_path=tempdir,
            files=[bw1, bw2],
            matrix_options=MatrixOptions(dtype=np.float32),
        )
"""

import os
import warnings
from multiprocessing import Pool
from typing import Union

import pandas as pd
from cellarr import buildutils_tiledb_frame as utf
from pyfaidx import Fasta

from . import build_options as bopt
from . import buildutils_tiledb_array as uta
from . import utils_bw as ubw
from .GenomicArrayDataset import GenomicArrayDataset

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


# TODO: Accept files as a dictionary with names to each dataset.
def build_genomicarray(
    files: list,
    output_path: str,
    features: Union[str, pd.DataFrame],
    genome_fasta: str,
    # genome: Union[str, pd.DataFrame] = "hg38",
    sample_metadata: Union[pd.DataFrame, str] = None,
    sample_metadata_options: bopt.SampleMetadataOptions = bopt.SampleMetadataOptions(),
    matrix_options: bopt.MatrixOptions = bopt.MatrixOptions(),
    feature_annotation_options: bopt.FeatureAnnotationOptions = bopt.FeatureAnnotationOptions(),
    optimize_tiledb: bool = True,
    num_threads: int = 1,
):
    """Generate the `GenomicArrayDatset`.

    All files are expected to be consistent and any modifications
    to make them consistent is outside the scope of this function
    and package.

    Args:
        files:
            List of file paths to `BigWig` files.

        output_path:
            Path to where the output TileDB files should be stored.

        features:
            A :py:class:`~pandas.DataFrame` containing the input genomic
            intervals..

            Alternatively, may provide path to the file containing a
            list of intervals. In this case,
            the first row is expected to contain the column names,
            "seqnames", "starts" and "ends".

            Additionally, the file may contain a column `sequences`,
            to specify the sequence string for each region. Otherwise,
            provide a link to the fasta file using the ``genome_fasta``
            parameter.

        genome_fasta:
            Path to a fasta file containing the sequence information.

            Sequence information will be updated for each region
            in ``features`` if the DataFrame/path does not contain
            a column `sequences`.

        sample_metadata:
            A :py:class:`~pandas.DataFrame` containing the sample
            metadata for each file in ``files``. Hences the number of rows
            in the dataframe must match the number of ``files``.

            Alternatively, may provide path to the file containing a
            concatenated sample metadata across all BigWig files. In this case,
            the first row is expected to contain the column names.

            Additionally, the order of rows is expected to be in the same
            order as the input list of ``files``.

            Defaults to `None`, in which case, we create a simple sample
            metadata dataframe containing the list of datasets, aka
            each BigWig files. Each dataset is named as ``sample_{i}``
            where `i` refers to the index position of the object in ``files``.

        sample_metadata_options:
            Optional parameters when generating ``sample_metadata`` store.

        matrix_options:
            Optional parameters when generating ``matrix`` store.

        feature_annotation_options:
            Optional parameters when generating ``feature_annotation`` store.

        optimize_tiledb:
            Whether to run TileDB's vaccum and consolidation (may take long).

        num_threads:
            Number of threads.
            Defaults to 1.
    """
    if not os.path.isdir(output_path):
        raise ValueError("'output_path' must be a directory.")

    ####
    ## Process genome information
    ####
    # if isinstance(genome, str):
    #     chrom_sizes = pd.read_csv(
    #         "https://hgdownload.soe.ucsc.edu/goldenpath/hg38/bigZips/hg38.chrom.sizes",
    #         sep="\t",
    #         header=None,
    #         names=["chrom", "length"],
    #     )
    # elif isinstance(genome, pd.DataFrame):
    #     chrom_sizes = genome
    #     if "chrom" not in chrom_sizes:
    #         raise ValueError("genome does not contain column: 'chrom'.")

    #     if "length" not in chrom_sizes:
    #         raise ValueError("genome does not contain column: 'length'.")

    # else:
    #     raise TypeError(
    #         "'genome' is not an expected type (either 'str' or 'Dataframe')."
    #     )

    ####
    ## Writing the features aka interval regions
    ####
    if isinstance(features, str):
        input_intervals = pd.read_csv(features, header=0)
    elif isinstance(features, pd.DataFrame):
        input_intervals = features.copy()

        required_cols = {"seqnames", "starts", "ends"}
        if not required_cols.issubset(input_intervals.columns):
            missing = required_cols - set(input_intervals.columns)
            raise ValueError(f"Missing required columns: {missing}")
    else:
        raise TypeError(
            "'input_intervals' is not an expected type (either 'str' or 'Dataframe')."
        )

    # append start index for each interval
    input_intervals["widths"] = input_intervals["ends"] - input_intervals["starts"]
    total_length = None
    if feature_annotation_options.aggregate_function is None:
        total_length = int(input_intervals["widths"].sum())
        counter = input_intervals["widths"].shift(1)
        counter[0] = 0
        input_intervals["genarr_feature_start_index"] = counter.cumsum().astype(int)
    else:
        counter = [1] * len(input_intervals)
        total_length = len(input_intervals)
        counter[0] = 0
        input_intervals["genarr_feature_start_index"] = (
            pd.Series(counter).cumsum().astype(int)
        )

    ends = input_intervals["genarr_feature_start_index"].shift(-1)
    ends.iloc[-1] = total_length
    input_intervals["genarr_feature_end_index"] = ends.astype(int)

    if not feature_annotation_options.skip:

        if "sequence" not in input_intervals.columns:
            gen_fa = Fasta(genome_fasta, as_raw=True)
            sequences = []
            for row in input_intervals.itertuples():
                seq = gen_fa.get_seq(row.seqnames, row.starts, row.ends)
                sequences.append(seq)

            input_intervals["sequences"] = sequences

        _col_types = utf.infer_column_types(
            input_intervals, feature_annotation_options.column_types
        )

        if "genarr_feature_index" not in input_intervals.columns:
            input_intervals["genarr_feature_index"] = range(0, len(input_intervals))

        _feature_output_uri = (
            f"{output_path}/{feature_annotation_options.tiledb_store_name}"
        )
        utf.create_tiledb_frame_from_dataframe(
            _feature_output_uri, input_intervals, column_types=_col_types
        )

        if optimize_tiledb:
            uta.optimize_tiledb_array(_feature_output_uri)

    ####
    ## Writing the sample metadata file
    ####
    _samples = []
    for idx, _ in enumerate(files):
        _samples.append(f"sample_{idx + 1}")

    if sample_metadata is None:
        warnings.warn(
            "Sample metadata is not provided, each dataset in 'files' is considered a sample",
            UserWarning,
        )

        sample_metadata = pd.DataFrame({"genarr_sample": _samples})
    elif isinstance(sample_metadata, str):
        sample_metadata = pd.read_csv(sample_metadata, header=0)
        if "genarr_sample" not in sample_metadata.columns:
            sample_metadata["genarr_sample"] = _samples
    elif isinstance(sample_metadata, pd.DataFrame):
        if "genarr_sample" not in sample_metadata.columns:
            sample_metadata["genarr_sample"] = _samples
    else:
        raise TypeError("'sample_metadata' is not an expected type.")

    if not sample_metadata_options.skip:
        _col_types = utf.infer_column_types(
            sample_metadata, sample_metadata_options.column_types
        )

        _sample_output_uri = (
            f"{output_path}/{sample_metadata_options.tiledb_store_name}"
        )
        utf.create_tiledb_frame_from_dataframe(
            _sample_output_uri, sample_metadata, column_types=_col_types
        )

        if optimize_tiledb:
            uta.optimize_tiledb_array(_sample_output_uri)

    ####
    ## Writing the genomic ranges file
    ####
    if not matrix_options.skip:
        _cov_uri = f"{output_path}/{matrix_options.tiledb_store_name}"
        uta.create_tiledb_array(
            _cov_uri,
            matrix_attr_name=matrix_options.matrix_attr_name,
            x_dim_dtype=feature_annotation_options.dtype,
            y_dim_dtype=sample_metadata_options.dtype,
            matrix_dim_dtype=matrix_options.dtype,
            x_dim_length=total_length,
            y_dim_length=len(files),
            is_sparse=False,
        )

        all_bws_options = [
            (
                _cov_uri,
                input_intervals,
                bwpath,
                idx,
                feature_annotation_options.aggregate_function,
                total_length,
            )
            for idx, bwpath in enumerate(files)
        ]

        if num_threads > 1:
            with Pool(num_threads) as p:
                p.map(_wrapper_extract_bwinfo, all_bws_options)
        else:
            for opt in all_bws_options:
                _wrapper_extract_bwinfo(opt)

        if optimize_tiledb:
            uta.optimize_tiledb_array(_cov_uri)

    return GenomicArrayDataset(
        dataset_path=output_path,
        sample_metadata_uri=sample_metadata_options.tiledb_store_name,
        feature_annotation_uri=feature_annotation_options.tiledb_store_name,
        matrix_tdb_uri=matrix_options.tiledb_store_name,
    )


def _write_intervals_to_tiledb(
    outpath, intervals, bwpath, bwidx, agg_func, total_length
):
    """Wrapper to extract the data for the given intervals from the bigwig file and write the output to the tiledb
    file."""
    data = ubw.wrapper_extract_bw_values(
        bw_path=bwpath,
        intervals=intervals,
        agg_func=agg_func,
        total_length=total_length,
    )

    if data is not None and len(data) > 0:
        uta.write_frame_intervals_to_tiledb(outpath, data=data, y_idx=bwidx)


def _wrapper_extract_bwinfo(args):
    """Wrapper for multiprocessing multiple files and intervals."""
    counts_uri, input_intervals, bwpath, idx, agg_func, total_length = args
    return _write_intervals_to_tiledb(
        counts_uri, input_intervals, bwpath, idx, agg_func, total_length
    )



import re
from functools import lru_cache
from typing import List, Union
from warnings import warn

import numpy as np
import pandas as pd
import tiledb

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


@lru_cache
def get_schema_names_frame(tiledb_obj: tiledb.Array) -> List[str]:
    """Get Attributes from a TileDB object.

    Args:
        tiledb_obj:
            A TileDB object.

    Returns:
        List of schema attributes.
    """
    columns = []
    for i in range(tiledb_obj.schema.nattr):
        columns.append(tiledb_obj.schema.attr(i).name)

    return columns


def subset_frame(
    tiledb_obj: tiledb.Array,
    subset: Union[slice, tiledb.QueryCondition],
    columns: list,
) -> pd.DataFrame:
    """Subset a TileDB object.

    Args:
        tiledb_obj:
            TileDB object to subset.

        subset:
            A :py:class:`slice` to subset.

            Alternatively, may provide a :py:class:`~tiledb.QueryCondition`
            to subset the object.

        columns:
            List specifying the atrributes from the schema to extract.

    Returns:
        A sliced `DataFrame` or a `matrix` with the subset.
    """

    if isinstance(subset, str):
        warn(
            "provided subset is string, its expected to be a 'query_condition'",
            UserWarning,
        )

        query = tiledb_obj.query(cond=subset, attrs=columns)
        data = query.df[:]
    else:
        data = tiledb_obj.df[subset][columns]

    re_null = re.compile(pattern="\x00")  # replace null strings with nan
    result = data.replace(regex=re_null, value=np.nan)
    result = result.dropna()
    return result


def _remap_index(indices: List[int]) -> List[int]:
    _map = {}
    _new_indices = []
    count = 0
    for r in list(indices):
        if r not in _map:
            _map[r] = count
            count += 1

        _new_indices.append(_map[r])

    return _new_indices, len(_map)


def subset_array(
    tiledb_obj: tiledb.Array,
    row_subset: Union[slice, list, tuple],
    column_subset: Union[slice, list, tuple],
    shape: tuple,
) -> np.ndarray:
    """Subset a tiledb storing array data.

    Uses multi_index to slice.

    Args:
        tiledb_obj:
            A TileDB object

        row_subset:
            Subset along the row axis.

        column_subset:
            Subset along the column axis.

        shape:
            Shape of the entire matrix.

    Returns:
        A dense array containing coverage.
    """
    data = tiledb_obj.multi_index[row_subset, column_subset]

    return data["data"]


def get_a_column(tiledb_obj: tiledb.Array, column_name: Union[str, List[str]]) -> list:
    """Access column(s) from the TileDB object.

    Args:
        tiledb_obj:
            A TileDB object.

        column_name:
            Name(s) of the column to access.

    Returns:
        List containing the column values.
    """
    if column_name not in get_schema_names_frame(tiledb_obj):
        raise ValueError(f"Column '{column_name}' does not exist.")

    if isinstance(column_name, str):
        column_name = [column_name]

    return tiledb_obj.query(attrs=column_name).df[:]


@lru_cache
def get_index(tiledb_obj: tiledb.Array) -> list:
    """Get the index of the TileDB object.

    Args:
        tiledb_obj:
            A TileDB object.

    Returns:
        A list containing the index values.
    """
    _index = tiledb_obj.unique_dim_values("__tiledb_rows")
    return [x.decode() for x in _index]


def _match_to_list(x: list, query: list):
    return sorted([x.index(q) for q in query])


def _is_list_strings(x: list):
    _ret = False

    if isinstance(x, (list, tuple)) and all(isinstance(y, str) for y in x):
        _ret = True

    return _ret



import tempfile

import numpy as np
import pandas as pd
import tiledb

from genomicarrays import build_genomicarray, FeatureAnnotationOptions

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_ingest_bigwigs():
    tempdir = tempfile.mkdtemp()

    strts = np.arange(300, 600, 20)
    features = pd.DataFrame(
        {"seqnames": ["chr1"] * 15, "starts": strts, "ends": strts + 15}
    )

    build_genomicarray(
        output_path=tempdir,
        files=["tests/data/test1.bw", "tests/data/test2.bw"],
        features=features,
        genome_fasta="tests/data/test.fa"
    )

    cfp = tiledb.open(f"{tempdir}/coverage", "r")
    ffp = tiledb.open(f"{tempdir}/feature_annotation", "r")

    features = ffp.df[:]

    assert len(features) == 15

    assert np.all(np.isnan(cfp.multi_index[:5, 0]["data"].flatten()))
    assert np.all(np.isnan(cfp.multi_index[:5, 1]["data"].flatten()))

    sfp = tiledb.open(f"{tempdir}/sample_metadata", "r")
    samples = sfp.df[:]
    assert len(samples) == 2

def test_ingest_bigwigs_agg():
    tempdir = tempfile.mkdtemp()

    strts = np.arange(300, 600, 20)
    features = pd.DataFrame(
        {"seqnames": ["chr1"] * 15, "starts": strts, "ends": strts + 15}
    )

    build_genomicarray(
        output_path=tempdir,
        files=["tests/data/test1.bw", "tests/data/test2.bw"],
        features=features,
        genome_fasta="tests/data/test.fa",
        feature_annotation_options=FeatureAnnotationOptions(aggregate_function=np.nanmean)
    )

    cfp = tiledb.open(f"{tempdir}/coverage", "r")
    ffp = tiledb.open(f"{tempdir}/feature_annotation", "r")

    features = ffp.df[:]

    assert len(features) == 15

    assert np.allclose(
        cfp.multi_index[0:5, 0]["data"],
        np.repeat(1, 5),
    )
    assert np.allclose(
        cfp.multi_index[0:5, 1]["data"],
        np.repeat(0.5, 5),
    )

    sfp = tiledb.open(f"{tempdir}/sample_metadata", "r")
    samples = sfp.df[:]
    assert len(samples) == 2



from dataclasses import dataclass
from typing import Dict, Literal, Optional, Callable

import numpy as np

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


@dataclass
class MatrixOptions:
    """Optional arguments for the ``matrix`` store for :py:func:`~genomicarrays.build_genomicarray.build_genomicarray`.

    Attributes:
        matrix_attr_name:
            Name of the matrix to be stored in the TileDB file.
            Defaults to "data".

        skip:
            Whether to skip generating matrix TileDB.
            Defaults to False.

        dtype:
            NumPy dtype for the values in the matrix.
            Defaults to np.uint16.

            Note: make sure the matrix values fit
            within the range limits of chosen-dtype.

        tiledb_store_name:
            Name of the TileDB file.
            Defaults to `coverage`.

        chunk_size:
            Size of chunks for parallel processing.

        compression:
            TileDB compression filter (None, 'gzip', 'zstd', 'lz4').

        compression_level:
            Compression level (1-9).
    """

    skip: bool = False
    matrix_attr_name: str = "data"
    dtype: np.dtype = np.float32
    tiledb_store_name: str = "coverage"
    chunk_size: int = 1000
    compression: Literal["zstd", "gzip", "lz4"] = "zstd"
    compression_level: int = 4

    def __post_init__(self):
        """Validate configuration."""
        if self.compression not in {"zstd", "gzip", "lz4", None}:
            raise ValueError(f"Unsupported compression: {self.compression}")

        if not 1 <= self.compression_level <= 9:
            raise ValueError(f"Invalid compression level: {self.compression_level}")


@dataclass
class SampleMetadataOptions:
    """Optional arguments for the ``sample`` store for :py:func:`~genomicarrays.build_genomicarray.build_genomicarray`.

    Attributes:
        skip:
            Whether to skip generating sample TileDB.
            Defaults to False.

        dtype:
            NumPy dtype for the sample dimension.
            Defaults to np.uint32.

            Note: make sure the number of samples fit
            within the integer limits of chosen dtype.

        tiledb_store_name:
            Name of the TileDB file.
            Defaults to "sample_metadata".

        column_types:
            A dictionary containing column names as keys
            and the value representing the type to in
            the TileDB.

            If `None`, all columns are cast as 'ascii'.
    """

    skip: bool = False
    dtype: np.dtype = np.uint32
    tiledb_store_name: str = "sample_metadata"
    column_types: Dict[str, np.dtype] = None


@dataclass
class FeatureAnnotationOptions:
    """Optional arguments for the ``feature`` store for :py:func:`~genomicarrays.build_genomicarray.build_genomicarray`.

    Attributes:
        skip:
            Whether to skip generating sample TileDB.
            Defaults to False.

        dtype:
            NumPy dtype for the sample dimension.
            Defaults to np.uint32.

            Note: make sure the number of features fit
            within the integer limits of chosen dtype.

        tiledb_store_name:
            Name of the TileDB file.
            Defaults to "feature_annotation".

        column_types:
            A dictionary containing column names as keys
            and the value representing the type to in
            the TileDB.

            If `None`, all columns are cast as 'ascii'.

        aggregate_function:
            A callable to summarize the values in a given
            interval. Defaults to None.
    """

    skip: bool = False
    dtype: np.dtype = np.uint32
    tiledb_store_name: str = "feature_annotation"
    column_types: Dict[str, np.dtype] = None
    aggregate_function: Optional[Callable] = None

    def __post_init__(self):
        if self.column_types is None:
            self.column_types = {"seqnames": "ascii", "starts": "int", "ends": "int"}



import numpy as np
import pyBigWig


def mock_bigwig(filename, max_range, value):
    bw = pyBigWig.open(filename, "w")
    bw.addHeader([("chr1", max_range)], maxZooms=0)

    starts = np.arange(0, max_range, 10, dtype=np.int64)
    ends = starts + 5
    values = np.array([value] * len(starts), dtype=np.float64)
    # np.array(np.random.random_sample(len(starts)), dtype=np.float64)
    chroms = np.array(["chr1"] * len(starts))

    bw.addEntries(chroms, starts, ends=ends, values=values)

    bw.close()


mock_bigwig("test1.bw", 1000, 1)
mock_bigwig("test2.bw", 500, 0.5)



"""Class that represents a realized subset of the `GenomicArrayDataset`.

This class provides a slice data class usually generated by the access
methods from
:py:func:`~genomicarray.GenomicArrayDataset.GenomicArrayDataset`.

Example:

    .. code-block:: python

        from genomicarray import GenomicArrayDataset

        garr = GenomicArrayDataset(dataset_path="/path/to/genomicarray/dir")
        feature_indices = slice(0, 10)
        result1 = garr[feature_indices, 0]

        print(result1)
"""

from dataclasses import dataclass
from typing import Any

import anndata
import genomicranges as gr
import pandas as pd
import summarizedexperiment as se

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


@dataclass
class GenomicArrayDatasetSlice:
    """Class that represents a realized subset of the `CellArrDataset`."""

    sample_metadata: pd.DataFrame
    feature_annotation: pd.DataFrame
    matrix: Any

    ####
    ## Interop
    ####

    def to_anndata(self):
        """Convert the realized slice to :py:class:`~anndata.AnnData`."""
        return anndata.AnnData(
            layers={"matrix": self.matrix.transpose()},
            obs=self.sample_metadata,
            var=self.feature_annotation,
        )

    def to_rangedsummarizedexperiment(self):
        """Convert the realized slice to
        :py:class:`~summarizedexperiment.RangedSummarizedExperiment.RangedSummarizedExperiment`."""
        return se.RangedSummarizedExperiment(
            assays={"matrix": self.matrix},
            row_ranges=gr.GenomicRanges.from_pandas(self.feature_annotation),
            column_data=self.sample_metadata,
        )

    ####
    ## Misc methods.
    ####

    @property
    def shape(self):
        return self.matrix.shape

    def __len__(self):
        return self.shape[0]

    ####
    ## Printing.
    ####

    def __repr__(self) -> str:
        """
        Returns:
            A string representation.
        """
        output = f"{type(self).__name__}(number_of_rows={self.shape[0]}"
        output += f", number_of_columns={self.shape[1]}"

        output += ")"
        return output

    def __str__(self) -> str:
        """
        Returns:
            A pretty-printed string containing the contents of this object.
        """
        output = f"class: {type(self).__name__}\n"

        output += f"number_of_rows: {self.shape[0]}\n"
        output += f"number_of_columns: {self.shape[1]}\n"

        return output
